from abc import ABC, abstractmethod

class ListOperations(ABC):
    def __init__(self):
        pass
    @abstractmethod
    def rpushKey(self, inMemoryReference, key, value):
        pass
    @abstractmethod
    def rpushKeyIfExists(self, inMemoryReference, key, value):
        pass
    @abstractmethod
    def lpushKey(self, inMemoryReference, key, value):
        pass
    @abstractmethod
    def llen(self, inMemoryReference, key):
        pass
    @abstractmethod
    def lpop(self, inMemoryReference, key):
        pass
    @abstractmethod
    def lindex(self, inMemoryReference, key,index):
        pass
    @abstractmethod
    def linsert(self, inMemoryReference,key, pivot, value):
        pass
    @abstractmethod
    def ltrim(self,inMemoryReference, key, start, stop):
        pass
    @abstractmethod
    def lset(self, inMemoryReference, key, index, value):
        pass
    @abstractmethod
    def rpop(self, inMemoryReference, key):
        pass

    @abstractmethod
    def lrange(self, inMemoryReference, key, start, end):
        pass

    @abstractmethod
    def createList(self, inMemoryReference, key):
        pass

    @abstractmethod
    def deleteList(self, inMemoryReference, key):
        pass



import string
from abc import ABC, abstractmethod
class StringOperations:
    def __init__(self):
        pass

    @abstractmethod
    def append(self, inMemoryDataStore, key, value):
        pass
    @abstractmethod
    def setKey(self, inMemoryDataStore, key, value):
        pass

    @abstractmethod
    def setKeyIfNotExists(self,inMemoryDataStore, key, value):
        pass

    @abstractmethod
    def strlenKey(self, inMemoryDataStore, key):
        pass

    @abstractmethod
    def getKey(self, inMemoryDataStore, key):
        pass

    @abstractmethod
    def incrKey(self, inMemoryDataStore, key):
        pass

    @abstractmethod
    def incrKeyBy(self, inMemoryDataStore, key,value):
        pass

    @abstractmethod
    def decrKey(self, inMemoryDataStore, key):
        pass

    @abstractmethod
    def decrKeyBy(self, inMemoryDataStore, key,value):
        pass

    @abstractmethod
    def delKey(self, inMemoryDataStore, key):
        pass

    @abstractmethod
    def exists(self, inMemoryDataStore, key):
        pass




import sys
from squirreldb.commands.stringOperations  import StringOperations

class StringOperationsImpl(StringOperations):
    def __init__(self):
        super().__init__()

    def append(self, inMemoryDataStore, key, value):
        if inMemoryDataStore == None or key == None or value ==None:
            return False
        if key in inMemoryDataStore and isinstance(inMemoryDataStore[key], str) and isinstance(value, str):
            inMemoryDataStore[key] = inMemoryDataStore[key]+value
            return True
        return False

    def setKey(self, inMemoryDataStore, key, value):
        if inMemoryDataStore == None or key == None or value == None:
            return False
        if isinstance(value, str) or isinstance(value, int):
            inMemoryDataStore[key] = value
            return True
        return False

    def setKeyIfNotExists(self,inMemoryDataStore, key, value):
        if inMemoryDataStore == None or key == None or value == None:
            return False
        if isinstance(value, str) or isinstance(value, int):
            if key not in inMemoryDataStore:
                inMemoryDataStore[key] = value
                return True
        return False

    def strlenKey(self, inMemoryDataStore, key):
        if inMemoryDataStore == None or key == None:
            return False
        if key in inMemoryDataStore and isinstance(inMemoryDataStore[key], str):
            return len(inMemoryDataStore[key])
        return False

    def getKey(self, inMemoryDataStore, key):
        if inMemoryDataStore == None or key == None:
            return False
        if key in inMemoryDataStore:
            return inMemoryDataStore[key]
        return False

    def incrKey(self, inMemoryDataStore, key):
        if inMemoryDataStore == None or key == None:
            return False
        if key not in inMemoryDataStore:
            inMemoryDataStore[key] = 0
        if isinstance(inMemoryDataStore[key], int) and inMemoryDataStore[key] < sys.maxsize-1:
            inMemoryDataStore[key] = inMemoryDataStore[key] + 1
            return True
        return False

    def decrKey(self, inMemoryDataStore, key):
        if inMemoryDataStore == None or key == None:
            return False
        if key not in inMemoryDataStore:
            inMemoryDataStore[key] = 0
        if isinstance(inMemoryDataStore[key], int) and inMemoryDataStore[key] > -sys.maxsize +1:
            inMemoryDataStore[key] = inMemoryDataStore[key] - 1
            return True
        return False

    def incrKeyBy(self, inMemoryDataStore, key,value):
        if inMemoryDataStore == None or value == None or not isinstance(value, int) or value < 0 or key == None:
            return False
        if key not in inMemoryDataStore:
            inMemoryDataStore[key] = 0
        if isinstance(inMemoryDataStore[key], int) and inMemoryDataStore[key] < sys.maxsize-value:
            inMemoryDataStore[key] = inMemoryDataStore[key] + value
            return True
        return False

    def decrKeyBy(self, inMemoryDataStore, key,value):
        if inMemoryDataStore == None or value == None or not isinstance(value, int) or value < 0 or key == None:
            return False
        if key not in inMemoryDataStore:
            inMemoryDataStore[key] = 0
        if isinstance(inMemoryDataStore[key], int) and inMemoryDataStore[key] > -sys.maxsize+value:
            inMemoryDataStore[key] = inMemoryDataStore[key] - value
            return True
        return False

    def delKey(self, inMemoryDataStore, key):
        if inMemoryDataStore == None or key == None or key not in inMemoryDataStore:
            return False
        del inMemoryDataStore[key]
        return True

    def exists(self, inMemoryDataStore, key):
        if inMemoryDataStore == None or key == None or key not in inMemoryDataStore:
            return False
        return True


from squirreldb.commands.stringOperationsImpl import StringOperationsImpl
from squirreldb.persistence.jsonPersistence import JsonPersistence
from squirreldb.engine.operations import Category
from squirreldb.persistence.storageState import StorageState
from squirreldb.commands.listOperationsImpl import ListOperationsImpl
import threading

class Squirrel():
    def __init__(self, filelocation=''):
        self.inMemoryStore = {}
        self.operations = {}
        self.operations[Category.STRING] = StringOperationsImpl()
        self.operations[Category.LIST] = ListOperationsImpl()
        self.persistence = JsonPersistence()
        self.databasefilename = ''
        # Lock Reference
        self.Thread = threading.Lock()
        self.databasefilelocation = filelocation
        if self.databasefilelocation != '' and self.databasefilename == '':
            self.databasefilename = self.databasefilelocation.split("/")[-1]
            self.persistence.load(self.databasefilelocation, self.inMemoryStore, StorageState.STALE)

    def DoThreadSafeOperation(self, callbackFunction, *argv):
        self.Thread.acquire()
        toret = False
        try:
            toret = callbackFunction(*argv)
        finally:
            self.Thread.release()
        return toret

    def get(self,key):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].getKey,self.inMemoryStore, key)

    def set(self, key, value):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].setKey,self.inMemoryStore,key,value)

    def append(self,key, value):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].append,self.inMemoryStore, key,value)

    def setIfNotExists(self, key, value):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].setKeyIfNotExists, self.inMemoryStore,key, value)

    def getStrLen(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].strlenKey, self.inMemoryStore, key)

    def incrKey(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].incrKey, self.inMemoryStore,key)

    def decrKey(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].decrKey, self.inMemoryStore,key)

    def decrKeyBy(self, key, value):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].decrKeyBy, self.inMemoryStore, key, value)

    def incrKeyBy(self,key, value):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].incrKeyBy, self.inMemoryStore, key, value)

    def delKey(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].delKey, self.inMemoryStore,key)

    def removeAll(self):
        if self.inMemoryStore == None:
            return False
        self.Thread.acquire()
        try:
            self.inMemoryStore.clear()
        finally:
            self.Thread.release()
        return True

    def exists(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.STRING].exists, self.inMemoryStore, key)

    def load(self, filepath, updatePolicy=StorageState.INMEMORY):
        return self.DoThreadSafeOperation(self.persistence.load, filepath,self.inMemoryStore, updatePolicy)

    def store(self, filepath):
        return self.DoThreadSafeOperation(self.persistence.store, self.inMemoryStore,filepath)

    def createList(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].createList, self.inMemoryStore,key)

    def deleteList(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].deleteList, self.inMemoryStore,key)

    def listRange(self, key, start, end):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].lrange, self.inMemoryStore,key, start,end)

    def listPopRight(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].rpop, self.inMemoryStore,key)

    def listSet(self, key, index, value):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].lset, self.inMemoryStore, key, index, value)

    def listTrim(self, key, start, end):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].ltrim, self.inMemoryStore,key, start, end)

    def listInsert(self, key, pivot, value):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].linsert, self.inMemoryStore, key, pivot, value)

    def listIndex(self, key, index):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].lindex, self.inMemoryStore, key, index)

    def listPopLeft(self,key):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].lpop, self.inMemoryStore, key)

    def listLength(self, key):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].llen, self.inMemoryStore, key)

    def listPushValueLeft(self, key, value):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].lpushKey, self.inMemoryStore, key, value)

    def listPushValueRight(self, key, value):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].rpushKey, self.inMemoryStore, key, value)

    def listPushValueRightIfExists(self, key, value):
        return self.DoThreadSafeOperation(self.operations[Category.LIST].rpushKeyIfExists, self.inMemoryStore, key, value)



from squirreldb.commands.listOperations import ListOperations

class ListOperationsImpl(ListOperations):

    def  __init__(self):
        super().__init__()

    def rpushKey(self, inMemoryReference, key, value):
        if inMemoryReference == None or key == None or value == None:
            return False
        if key not in inMemoryReference:
            inMemoryReference[key] = []
        if isinstance(inMemoryReference[key], list):
            if len(inMemoryReference[key]) == 0:
                inMemoryReference[key].append(value)
                return True
            elif type(inMemoryReference[key][-1]) == type(value):
                inMemoryReference[key].append(value)
                return True
        return False

    def rpushKeyIfExists(self, inMemoryReference, key, value):
        if inMemoryReference == None or key == None or value == None:
            return False
        if key in inMemoryReference:
            return self.rpushKey(inMemoryReference, key, value)
        else:
            return False

    def lpushKey(self, inMemoryReference, key, value):
        if inMemoryReference == None or key == None or value == None:
            return False
        if key not in inMemoryReference:
            inMemoryReference[key] = []
        if isinstance(inMemoryReference[key], list):
            if len(inMemoryReference[key]) == 0:
                inMemoryReference[key].insert(0,value)
                return True
            elif type(inMemoryReference[key][0]) == type(value):
                inMemoryReference[key].insert(0, value)
                return True
        return False

    def llen(self, inMemoryReference, key):
        if inMemoryReference == None or key == None:
            return False
        if key in inMemoryReference:
            if isinstance(inMemoryReference[key],list):
                return len(inMemoryReference[key])
        return False

    def lpop(self, inMemoryReference, key):
        if inMemoryReference == None or key == None:
            return False
        if key in inMemoryReference:
            if isinstance(inMemoryReference[key], list):
                if len(inMemoryReference[key]) != 0:
                    return inMemoryReference[key].pop(0)
        return False

    def lindex(self, inMemoryReference, key,index):
        if inMemoryReference == None or key == None or index == None:
            return False
        if key in inMemoryReference:
            if isinstance(inMemoryReference[key], list):
                inMemoryListReference = inMemoryReference[key]
                #Pretty compilcated white listing of the index value
                if type(index) is int and index > 0 and index < len(inMemoryListReference):
                    return inMemoryListReference[index]
        return False


    def linsert(self, inMemoryReference, key, pivot, value):
        if inMemoryReference == None or key == None or pivot == None or value == None:
            return  False
        if key in inMemoryReference:
            inMemoryListReference = inMemoryReference[key]
            if isinstance(inMemoryListReference, list):
                if isinstance(pivot, int):
                    if 0 < pivot < len(inMemoryListReference):
                        if type(inMemoryListReference[-1]) == type(value):
                            inMemoryListReference.insert(pivot, value)
                            return True
                    elif pivot == 0:
                        if len(inMemoryListReference) > 0 and type(inMemoryListReference[-1]) == type(value):
                            inMemoryListReference.insert(pivot,value)
                            return True
                        elif len(inMemoryListReference) == 0:
                            inMemoryListReference.append(value)
                            return True
                    elif pivot >= len(inMemoryListReference):
                        if len(inMemoryListReference) > 0 and type(inMemoryListReference[-1]) == type(value):
                            inMemoryListReference.append(value)
                            return True
                        else:
                            inMemoryListReference.append(value)
                            return True
        return False


    def ltrim(self, inMemoryReference, key, start, stop):
        if inMemoryReference == None or key == None or start == None or stop == None:
            return False
        if not isinstance(start, int) or not isinstance(stop, int) or start > stop or start < 0:
            return False
        if key in inMemoryReference:
            inMemoryListReference = inMemoryReference[key]
            if isinstance(inMemoryListReference, list):
                if start < len(inMemoryListReference):
                    inMemoryReference[key] = inMemoryListReference[start:stop+1]
                    return True
                elif start >= len(inMemoryListReference):
                    inMemoryReference[key] = []
                    return True
        return False

    def lset(self, inMemoryReference, key, index, value):
        if inMemoryReference == None or key == None or index == None or value == None:
            return False
        if key in inMemoryReference:
            inMemoryListReference = inMemoryReference[key]
            if isinstance(inMemoryListReference, list):
                if isinstance(index, int) and 0 <= index < len(inMemoryListReference):
                    if type(inMemoryListReference[index]) == type(value):
                        inMemoryListReference[index] = value
                        return True
                    if len(inMemoryListReference) == 0 and index == 0:
                        inMemoryListReference.append(value)
                        return True
        elif isinstance(index, int) and index == 0:
            inMemoryReference[key] = [value]
            return True
        return False


    def rpop(self, inMemoryReference, key):
        if inMemoryReference == None or key == None:
            return False
        if key in inMemoryReference:
            inMemoryListReference = inMemoryReference[key]
            if isinstance(inMemoryListReference, list):
                if len(inMemoryListReference) > 0:
                    return inMemoryListReference.pop(-1)
        return False


    def lrange(self, inMemoryReference, key, start, end):
        if inMemoryReference == None or key == None or start == None or end == None:
            return False
        if not isinstance(start, int) or not isinstance(end, int) or start > end or start < 0:
            return False
        if key in inMemoryReference:
            inMemoryListReference = inMemoryReference[key]
            if isinstance(inMemoryListReference , list):
                if start < len(inMemoryListReference):
                    return inMemoryListReference[start:end+1]
                elif start >= len(inMemoryListReference):
                    return []
        return False

    def createList(self, inMemoryReference, key):
        if inMemoryReference == None or key == None:
            return False
        if key not in inMemoryReference:
            inMemoryReference[key] = []
            return True
        return False


    def deleteList(self, inMemoryReference, key):
        if inMemoryReference == None or key == None:
            return False
        if key in inMemoryReference:
            if isinstance(inMemoryReference[key], list):
                del inMemoryReference[key]
                return True
        return False



import json
import os
import datetime
from squirreldb.persistence.persistence import Persistence
from squirreldb.persistence.storageState import StorageState
class JsonPersistence(Persistence):
    def __init__(self):
        super().__init__()
        self.filename = ""
        self.filePath = ""
        self.timeofupdation = ""
        self.timeofcreation = ""
    """
        Loads database dump stored in file into inMemory datastore. 
        Always loads any key that's not there in current InMemory datastore.
        If updatePolicy is to STALE, any key value in inMemoryDatastore is updated with Stored file data. 
        On INMEMORY updatePolicy, any data in InMemory is given higher preference.
        
        A User can call load either at start of his API interaction or Inbetween his inter-actions.
    """
    def load(self, databaseFilePath, inMemoryReference, updatePolicy=StorageState.INMEMORY):
        if databaseFilePath == None or inMemoryReference == None:
            return False
        try:
            with open(databaseFilePath, "r") as databaseFile:
                tempReference = json.load(databaseFile)
                for key in tempReference:
                    if key in inMemoryReference and updatePolicy == StorageState.STALE:
                        inMemoryReference[key] = tempReference[key]
                    if key not in inMemoryReference:
                        inMemoryReference[key] = tempReference[key]
        except FileNotFoundError:
            print("Information: Database file not found, init empty database")
            if updatePolicy == StorageState.STALE:
                inMemoryReference.clear()
            return  True
        except ValueError:
            print("Error: , Database file corrupt / in-unreadable format")
            return  False
        return True

    """
        Creates Store file, serializes any data in inMemoryStore and writes it off to given file path. 
        If file already exists, this operation basically overwrites the stale data. So caution must be taken if the existing store file may be required
        for future audit.
    """

    def store(self, inMemoryReference, databaseFilePath):
        if inMemoryReference == None or databaseFilePath == None:
            return False
        try:
            databaseFileExists = os.path.isfile(databaseFilePath)
            databaseDumpFile = open(databaseFilePath, "w")
            if databaseFileExists:
                databaseDumpFile.truncate(0)
            else:
                self.timeofcreation = datetime.datetime.now()
            self.timeofupdation = datetime.datetime.now()
            json.dump(inMemoryReference, databaseDumpFile)
        except Exception:
            print("Value Error in storing data ")
            return False
        return True

    def close(self):
        self.timeofcreation = ""
        self.timeofupdation = ""
        self.filename =""
        self.filePath = ""
        return True





from abc import ABC, abstractmethod
class Persistence(ABC):
    def __init__(self):
        self.filename = ""
        self.filePath = ""
        self.timeofcreation = ""
        self.timeofupdation = ""
        pass
    @abstractmethod
    def load(self, databaseFilePath, inMemoryReference):
        pass
    @abstractmethod
    def store(self, inMemoryReference, databaseFilePath):
        pass
    @abstractmethod
    def close(self):
        pass



import setuptools

setuptools.setup(
    name="SquirrelDB",
    version='0.2.5',
    author="Vigneash Sundararajan",
    author_email = "vigneashsundar@live.com",
    description="Type & Thread safe In-Memory concurrent datastore written in Python.",
    packages=setuptools.find_packages(),
    license = 'GPL',
    long_description=open('README.MD').read(),
    url = "https://github.com/vikene/squirreldb",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)
